{
  SensorDriver.spin

  Summary:
                This object provides data from ITG3200, ADXL345, and HMC5883L sensors.
                Data is update at a rate of 200Hz for the ITG3200 and ADXL345 and 50Hz for the HMC5883L

  Usage:
                StartSensorData(COG, SCL, SDA, GyroDLPF)
                                This starts the driver, defines the SCL and SDA pins and the low pass
                                frequency for the ITG3200 sensor
                                
                
                GetGyroX / GetGyroY / GetGyroZ
                GetAccX / GetAccY / GetAccZ
                GetMagX / GetMagY / GetMagZ
                                These functions return the values from the various sensors
  
  
}


CON
        ' Data indexes
        TEMP      =             0

        GYRO_X    =             1
        GYRO_Y    =             2
        GYRO_Z    =             3

        ACC_X     =             4
        ACC_Y     =             5
        ACC_Z     =             6

        MAG_X     =             7
        MAG_Y     =             8
        MAG_Z     =             9

        

        ' General Constants
        DEV_ITG3200             = 1
        DEV_ADXL345             = 2
        DEV_HMC5883L            = 3


        ' I2C bus constants

        ACK                     = $0
        NAK                     = $1

        DELAY_CYCLES            = 52           ' I2C Delay time.  Must be between 12 and 511

        PASS                    = $1
        FAIL                    = $0


        ' Commands
        CMD_INIT                = 1
        CMD_READ_REG            = 2
        CMD_SET_REG             = 3
        CMD_ITG_UPDATE_DATA     = 4
        CMD_ADXL_UPDATE_DATA    = 5
        CMD_HMC_UPDATE_DATA     = 6
        CMD_AUTO_UPDATE_START   = 7
        CMD_AUTO_UPDATE_STOP    = 8


        AUTO_DELAY              = 800_000
 
        ' I2C Device Addresses
        ITG3200_RdAddr            = %11010001
        ITG3200_WrAddr            = %11010000
        
        ADXL345_RdAddr            = %00111011
        ADXL345_WrAddr            = %00111010
        
        HMC5883L_RdAddr           = %00111101
        HMC5883L_WrAddr           = %00111100



        
        ' ===============ITG3200 Configuration====================
                      
        ' Sample Rate Divider
        ' Fs = Fi / (divider + 1)  where Fi is either 1kHz or 8kHz
        ' Valid values are from $00 to $FF
        ITG_SMPLRT_DIV              = $04

        
        ' Full Scale Select
        ' Must be set to $03 (+/- 2000 degrees/sec)
        ITG_FS_SEL                  = $03

        
        ' Digital low pass filter configuration
        ' 0 => 256Hz
        ' 1 => 188Hz
        ' 2 => 98Hz
        ' 3 => 42Hz
        ' 4 =>  20Hz
        ' 5 =>  10Hz
        ' 6 =>  5Hz

        ITG_DLPF_256HZ          = $00
        ITG_DLPF_188HZ          = $01
        ITG_DLPF_98HZ           = $02
        ITG_DLPF_42HZ           = $03
        ITG_DLPF_20HZ           = $04
        ITG_DLPF_10HZ           = $05
        ITG_DLPF_5HZ            = $06

        
        ' Clock Select
        ' 0 => Internal Oscillator
        ' 1 => PLL with X Gyro ref
        ' 2 => PLL with Y Gyro ref
        ' 3 => PLL with Z Gyro ref
        ' 4 => PLL with external 32.768kz ref,
        ' 5 => PLL with external 19.2Mhz ref
        ITG_CLK_SEL                 = $01


        ' Interrupt Configuration
        ' bit 7 : 1=active low, 0=active hight
        ' bit 6 : 1=open drain, 0=push-pull
        ' bit 5 : 1=latch till clear, 0=50us pulse
        ' bit 4 : 1=clear on any read, 0=clear on status register read
        ' bit 3 : 0
        ' bit 2 : enable interrupt when device is ready (PLL ready after change clock)
        ' bit 1 : 0
        ' bit 0 : enable interrupt when data is available
        ITG_INT_CFG                 = %00110001
        '(active low, push-pull, latch till clear, clear on any read, interrupt on new data) 
        

        ' Register Map
        ITG_RegWHOAMI              = $00
        ITG_RegSMPLRT_DIV          = $15
        ITG_RegDLPF_FS             = $16
        ITG_RegINT_CFG             = $17
        ITG_RegINT_STATUS          = $1A
        ITG_RegTEMP_OUT_H          = $1B        
        ITG_RegTEMP_OUT_L          = $1C        
        ITG_RegGYRO_XOUT_H         = $1D        
        ITG_RegGYRO_XOUT_L         = $1E        
        ITG_RegGYRO_YOUT_H         = $1F        
        ITG_RegGYRO_YOUT_L         = $20        
        ITG_RegGYRO_ZOUT_H         = $21        
        ITG_RegGYRO_ZOUT_L         = $22
        ITG_RegPWR_MGM             = $3E




        
        ' ============ADXL345 Configuration===================                


        ADXL_DeviceID        =  $E5  ' Fixed DeviceID
  

        ADXL_Data_FormatStr  =  %0000_1001 ' D7 = 0 Self-test disabled
                                           ' D6 = 0 SPI set to 4-wire mode
                                           ' D5 = 0 Interupts active high
                                           ' D4 = 0 No function
                                           ' D3 = 1 Full resolution mode
                                           ' D2 = 0 righ justified mode with sign extension
                                           ' D1 = 0 Range Setting (Table 18) +/- 4g
                                           ' D0 = 1 Range Setting (Table 18) +/- 4g   

        ADXL_POWER_CTLStr    =  %0000_1000 ' D7 = 0 No function
                                           ' D6 = 0 No function
                                           ' D5 = 0 Activity and inactivity concurrent (see Link Mode section)
                                           ' D4 = 0 No Sleep when inactivity is detected
                                           ' D3 = 1 Measurement mode ON
                                           ' D2 = 0 Sleep OFF Normal mode ON
                                           ' D1 = 0 Frequency of readings in Sleep mode 8Hz
                                           ' D0 = 0 Frequency of readings in Sleep mode 8Hz
                                                         
        ADXL_BW_RATEStr      =  %0000_1011 ' D7 = 0 No function
                                           ' D6 = 0 No function
                                           ' D5 = 0 No function
                                           ' D4 = 0 Normal operation
                                           ' D3 = 1 Data Rate Setting Output 3200Hz Bandwidth 1600Hz (Table 6)
                                           ' D2 = 1 Data Rate Setting Output 3200Hz Bandwidth 1600Hz (Table 6)
                                           ' D1 = 1 Data Rate Setting Output 3200Hz Bandwidth 1600Hz (Table 6)
                                           ' D0 = 1 Data Rate Setting Output 3200Hz Bandwidth 1600Hz (Table 6)

  

        ' Register Map
        ADXL_RegDEVID        =  $00  ' Device ID fixed as $E5
        ADXL_RegOFSX         =  $1E  ' User-set offset adjustments in 2's complement 15.6mg/LSB ($7F = +2g)
        ADXL_RegOFSY         =  $1F  ' User-set offset adjustments in 2's complement 15.6mg/LSB ($7F = +2g)
        ADXL_RegOFSZ         =  $20  ' User-set offset adjustments in 2's complement 15.6mg/LSB ($7F = +2g)
        ADXL_RegBW_RATE      =  $2C  ' Sets power mode and device bandwidth and output data rate Table 6&7 (default 100Hz)
        ADXL_RegPOWER_CTL    =  $2D  ' Sets Sleep and Wakeup used for activity function
        ADXL_RegINT_ENABLE   =  $2E  ' Enables interupts.  Config before enabling. Always enabled
        ADXL_RegINT_MAP      =  $2F  ' Maps interupt functions t interupt pins.
        ADXL_RegDATA_FORMAT  =  $31  ' Controls presentation of data on RegDATA** registers
        ADXL_RegDATAX0       =  $32  ' Data in 2's complement with "0" LSByte and "1" MSByte
        ADXL_RegDATAX1       =  $33  ' Data in 2's complement with "0" LSByte and "1" MSByte 
        ADXL_RegDATAY0       =  $34  ' Data in 2's complement with "0" LSByte and "1" MSByte 
        ADXL_RegDATAY1       =  $35  ' Data in 2's complement with "0" LSByte and "1" MSByte 
        ADXL_RegDATAZ0       =  $36  ' Data in 2's complement with "0" LSByte and "1" MSByte 
        ADXL_RegDATAZ1       =  $37  ' Data in 2's complement with "0" LSByte and "1" MSByte         
        ADXL_RegFIFO_CTL     =  $38  ' Sets FIFO mode
        ADXL_RegFIFO_STATUS  =  $39  ' Sets up to 32 entries and shows FIFO_TRIG value




        

        ' =============HMC5883L Configuration==================

        ' Configuration settings
        HMC_ConfigAStr          = %0_00_110_00          ' D7 = No function
                                                        ' D6-D5 = 00 (1 sample per output)
                                                        ' D4-D2 = 110 (75 Hz output)
                                                        ' D1-D0 = 00 (No measurement bias)
                                                        
        HMC_ConfigBStr          = %001_00000            ' D7-D5 = 001 (+/- 1.5 Ga)
                                                        ' D4-D0 = No function
        
        HMC_ModeStr             = %000000_00            ' D7-D2 = No function
                                                        ' D1-D0 = 00 (Continuous conversion)

        
        ' Register Map
        HMC_RegConfigA          = $00
        HMC_RegConfigB          = $01
        HMC_RegMode             = $02
        HMC_RegXHi              = $03
        HMC_RegXLo              = $04
        HMC_RegZHi              = $05
        HMC_RegZLo              = $06
        HMC_RegYHi              = $07
        HMC_RegYLo              = $08
        HMC_RegStatus           = $09
        HMC_RegIDA              = $0A
        HMC_RegIDB              = $0B
        HMC_RegIDC              = $0C


OBJ
     

VAR
  ' This holds all data values transferred between the
  ' spin code and the asm code
  long dataValues[10]

  long buffer
  long  xIndex, yIndex, zIndex


PUB StartSensorData(_cog, _sclPin, _sdaPin, _gyroDLPF)

  sclPin := 1 << _sclPin
  sdaPin := 1 << _sdaPin

  itgDLPF := _gyroDLPF
 
  coginit(_cog, @DataRun, @dataValues)

' Returns the data value at the specified index
PUB GetValue(index)
  return dataValues[index]


PUB GetGyroTemp
  return dataValues[TEMP]

  
' Returns the X gyro value corrected for bias
PUB GetGyroX | value
  return (dataValues[GYRO_X])

' Returns the Y gyro value corrected for bias
PUB GetGyroY | value
  return (dataValues[GYRO_Y])

' Returns the Z gyro value corrected for bias
PUB GetGyroZ | value
  return (dataValues[GYRO_Z])

' Returns the X acc value
PUB GetAccX | value
  return dataValues[ACC_X]


' Returns the Y acc value 
PUB GetAccY | value
  return dataValues[ACC_Y]


' Returns the Z acc value
PUB GetAccZ | value
  return dataValues[ACC_Z]


' Returns the X mag value
PUB GetMagX | t
  return dataValues[MAG_X]
  
' Returns the Y mag value
PUB GetMagY | t
  return dataValues[MAG_Y]


' Returns the Z mag value
PUB GetMagZ | t
  return dataValues[MAG_Z]




DAT
' Assembly engine
        org   0

DataRun
              call      #init

              mov       magCnt, #4
              mov       loopCount, cnt
              add       loopCount, loopDelay
:outerLoop
              call      #ITGdataUpdate
              'call      #ITGComputeDrift
              call      #ADXLdataUpdate

              ' Only update magnetometer every 4 cycles (50hz)
              djnz      magCnt, #:skipMagUpdate
              call      #HMCdataUpdate
              mov       magCnt, #4
:skipMagUpdate

              waitcnt   loopCount, loopDelay               
              jmp       #:outerLoop
              
DataRun_ret ret



init          ' Set the SCL / SDA Pin


              mov       writeAddr, #ITG3200_wrAddr
              mov       readAddr, #ITG3200_rdAddr
              
              ' Send up to 9 clock cycles and look
              ' for a NAK response
              or        outa, sclPin
              or        dira, sclPin
              andn      dira, sdaPin
              call      #delay
              mov       counter, #9
:loop         andn      outa, sclPin
              call      #delay
              call      #delay
              or        outa, sclPin
              call      #delay
              test      sdaPin, ina wz 
        if_nz jmp       #:pass
              djnz      counter, #:loop

              ' If we get here, we never saw a NAK
              mov       parm, #FAIL
              jmp       #init_ret

:pass
              ' Set the sample rate divider
              mov       regAddr, #ITG_RegSMPLRT_DIV
              mov       regData, #ITG_SMPLRT_DIV
              call      #setReg

              
              ' Set the DLPF_FS register
              mov       regAddr, #ITG_RegDLPF_FS
              mov       regData, #ITG_FS_SEL
              shl       regData, #3
              or        regData, itgDLPF '#ITG_DLPF_CFG
              call      #setReg

              
              ' Set the Power Management register
              mov       regAddr, #ITG_RegPWR_MGM
              mov       regData, #ITG_CLK_SEL
              call      #setReg


              ' Set the Interrupt register
              mov       regAddr, #ITG_RegINT_CFG
              mov       regData, #ITG_INT_CFG
              call      #setReg



              
              ' CONFIGURE THE ADXL345
              mov       writeAddr, #ADXL345_wrAddr
              mov       readAddr, #ADXL345_rdAddr


              ' Data Format
              mov       regAddr, #ADXL_RegDATA_FORMAT
              mov       regData, #ADXL_Data_FormatStr
              call      #setReg

              ' Data Rate
              mov       regAddr, #ADXL_RegBW_RATE
              mov       regData, #ADXL_BW_RATEStr
              call      #setReg

              ' Power Control
              mov       regAddr, #ADXL_RegPOWER_CTL
              mov       regData, #ADXL_POWER_CTLStr
              call      #setReg              



              ' CONFIGURE THE HMC5883L
              mov       writeAddr, #HMC5883L_wrAddr
              mov       readAddr, #HMC5883L_rdAddr

              ' Config A
              mov       regAddr, #HMC_RegConfigA
              mov       regData, #HMC_ConfigAStr
              call      #setReg
              
              ' Config B
              mov       regAddr, #HMC_RegConfigB
              mov       regData, #HMC_ConfigBStr
              call      #setReg

              ' Mode
              mov       regAddr, #HMC_RegMode
              mov       regData, #HMC_ModeStr
              call      #setReg
              
              ' set pass condition
              mov       parm, #PASS                     
              
init_ret      ret




ITGdataUpdate
              mov       devType, #DEV_ITG3200
              call      #setDevice

              ' Set the hubAddr to the first output (temperature)
              mov       hubAddr, par
              
              call      #start

              ' Send the write address
              mov       parm, writeAddr
              call      #writebyte

              ' Send the register to be read
              mov       parm, #ITG_RegTEMP_OUT_H
              call      #writebyte

              call      #start

              ' Send the read address
              mov       parm, readAddr
              call      #writebyte

              ' Loop four times (temp, x, y, z)
              mov       updateCounter, #4
:loop         mov       parm, #ACK
              call      #readbyte

              ' Check sign for sign extension
              mov       inData, #0
              test      parm, #$80              wz
        if_nz sub       inData, #$100

              ' Save lower byte
              or        inData, parm
              shl       inData, #8

              ' Send ACK unless we are on last loop
              mov       parm, #ACK
              cmp       updateCounter, #1       wz
        if_z  mov       parm, #NAK
              call      #readbyte
              or        inData, parm

              ' Write out data and increment the hubAddr
     '         subs      inData, drift                   ' Account for drift
              wrlong    inData, hubAddr
              add       hubAddr, #4

              djnz      updateCounter, #:loop

              call      #stop


ITGdataUpdate_ret ret




ITGComputeDrift
                        mov     drift, iT               ' Start with the temperature reading
                        add     drift, tempOffset       ' Offset it by 15,000

                        ' divide drift by 100                        

                        mov     divisor, #100
                        mov     dividend, drift
                        test    dividend, i2cWordReadMask    wc

                        muxc    signbit, #1             ' record the sign of the original value
                        abs     dividend, dividend

                        mov     divCounter, #10     
                        shl     divisor, divCounter
                        mov     resultShifted, #1
                        shl     resultShifted, divCounter

                        add     divCounter, #1
                        mov     drift, #0

:divLoop                        
                        cmp     dividend, divisor   wc
              if_nc     add     drift, resultShifted
              if_nc     sub     dividend, divisor
                        shr     resultShifted, #1
                        shr     divisor, #1     
                        djnz    divCounter, #:divLoop

                        test    signbit, #1     wc
                        negc    drift, drift

                        
ITGComputeDrift_Ret     ret





ADXLdataUpdate

              mov       devType, #DEV_ADXL345
              call      #setDevice

              ' Set the hubAddr to the first acc output (index 4)
              mov       hubAddr, par
              add       hubAddr, #16
              
              call      #start

              ' Send the write address
              mov       parm, writeAddr
              call      #writebyte

              ' Send the register to be read
              mov       parm, #ADXL_RegDATAX0
              call      #writebyte

              call      #start

              ' Send the read address
              mov       parm, readAddr
              call      #writebyte

              ' Loop three times (x, y, z)
              mov       updateCounter, #3
:loop         mov       parm, #ACK
              call      #readbyte

              ' Grab the low byte
              mov       loByte, parm

              ' Send ACK unless we are on last loop
              mov       parm, #ACK
              cmp       updateCounter, #1       wz
        if_z  mov       parm, #NAK
              call      #readbyte

              ' Grab the high byte
              mov       hiByte, parm
                  
              ' Check sign for sign extension
              mov       inData, #0
              test      hiByte, #$80              wz
        if_nz sub       inData, #$100

              ' Save high byte
              or        inData, hiByte
              shl       inData, #8

              ' Save the low byte
              or        inData, loByte

              ' Write out data and increment the hubAddr
              wrlong    inData, hubAddr
              add       hubAddr, #4

              djnz      updateCounter, #:loop

              call      #stop

ADXLdataUpdate_ret ret



HMCdataUpdate
              mov       devType, #DEV_HMC5883L
              call      #setDevice

              ' Set the hubAddr to the first output (X Mag)
              mov       hubAddr, par
              add       hubAddr, #28

              
              call      #start

              ' Send the write address
              mov       parm, writeAddr
              call      #writebyte

              ' Send the register to be read
              mov       parm, #HMC_RegXHi
              call      #writebyte

              call      #start

              ' Send the read address
              mov       parm, readAddr
              call      #writebyte

              ' Loop three times (temp, x, z, y)
              mov       updateCounter, #3
:loop         mov       parm, #ACK
              call      #readbyte

              ' Check sign for sign extension
              mov       inData, #0
              test      parm, #$80              wz
        if_nz sub       inData, #$100

              ' Save higher byte
              or        inData, parm
              shl       inData, #8

              ' Send ACK unless we are on last loop
              mov       parm, #ACK
              cmp       updateCounter, #1       wz
        if_z  mov       parm, #NAK

              ' Get the lower byte
              call      #readbyte
              or        inData, parm

              ' Correct for X, Z, Y ordering on HMC5883L
              mov       hubAddr, par              
              cmp      updateCounter, #3       wz
        if_z  add       hubAddr, #32
        if_z  neg       inData, inData
              cmp       updateCounter, #2       wz
        if_z  add       hubAddr, #36
              cmp       updateCounter, #1       wz
        if_z  add       hubAddr, #28
        
              ' Write out data 
              wrlong    inData, hubAddr

              djnz      updateCounter, #:loop

              call      #stop

              ' Trigger the next conversion
              mov       regAddr, #HMC_RegMode
              mov       regData, #HMC_ModeStr
              call      #setReg
              

HMCdataUpdate_ret ret
                 

setDevice
              cmp       devType, #DEV_ITG3200   wz
        if_z  mov       writeAddr, #ITG3200_wrAddr
        if_z  mov       readAddr,#ITG3200_rdAddr

              cmp       devType, #DEV_ADXL345   wz
        if_z  mov       writeAddr, #ADXL345_wrAddr
        if_z  mov       readAddr, #ADXL345_rdAddr

              cmp       devType, #DEV_HMC5883L   wz
        if_z  mov       writeAddr, #HMC5883L_wrAddr
        if_z  mov       readAddr, #HMC5883L_rdAddr

setDevice_ret ret


setReg
              call      #start

              ' Send the write address
              mov       parm, writeAddr
              call      #writebyte

              ' Send the srd register
              mov       parm, regAddr
              call      #writebyte

              ' Send the new value
              mov       parm, regData
              call      #writebyte

              call      #stop

setReg_ret    ret             


start         or        outa, sdaPin            ' Set SDA HIGH
              or        dira, sdaPin            ' Set SDA as output
              call      #delay
              or        outa, sclPin            ' Set SCL HIGH
              or        dira, sclPin            ' Set SCL as output
              call      #delay
              andn      outa, sdaPin            ' Set SDA LOW
              call      #delay
              andn      outa, sclPin            ' Set SCL LOW
              call      #delay
start_ret     ret



stop          andn      outa, sdaPin            ' Set SDA LOW
              call      #delay
              or        outa, sclPin            ' Set SCL HIGH
              call      #delay
              or        outa, sdaPin            ' Set SDA HIGH
              call      #delay
              andn      dira, sclPin            ' Float SCL HIGH
              andn      dira, sdaPin            ' Float SDA HIGH
stop_ret      ret



writebyte     mov       data, parm              ' Get the data byte
              mov       counter, #8             ' Set loop count for 8 bits

:loop         shl       data, #1                ' Shift left one bit
              test      data, #$100    wz       ' Check MSB
        if_z  andn      outa, sdaPin            ' Set SDA LOW if zero
        if_nz or        outa, sdaPin            ' Set SDA HIGH if not zero
              call      #delay
              or        outa, sclPin            ' Set SCL HIGH
              call      #delay
              andn      outa, sclPin            ' Set SCL LOW
              call      #delay
              djnz      counter, #:loop

              andn      dira, sdaPin            ' Set SDA as input
              call      #delay
              or        outa, sclPin            ' Set SCL HIGH
              call      #delay
              test      sdaPin, ina    wz       ' Check SDA input
        if_z  mov       ackbit, #0              ' Set to zero if LOW
        if_nz mov       ackbit, #1              ' Set to one if HIGH
              andn      outa, sclPin            ' Set SCL LOW
              call      #delay
              or        dira, sdaPin            ' Set SDA as output
              mov       parm, ackbit            ' Return the ack bit
writebyte_ret ret




readbyte      mov       ackbit, parm            ' Get the ACK bit
              mov       data, #0                ' Initialize data byte to zero
              andn      dira, sdaPin            ' Set SDA as input
              call      #delay
              mov       counter, #8             ' Set loop count for 8

:loop         call      #delay
              or        outa, sclPin            ' Set SCL HIGH
              call      #delay
              shl       data, #1                ' data byte left one bit
              test      sdaPin, ina    wz
        if_nz or        data, #1                ' Set LSB if input bit is HIGH
              andn      outa, sclPin            ' Set SCL LOW
              call      #delay
              djnz      counter, #:loop

              cmp       ackbit, #0     wz
        if_z  andn      outa, sdaPin            ' Set SDA LOW if ACK
        if_nz or        outa, sdaPin            ' Set SDA HIGH if NAK
              or        dira, sdaPin            ' Set SDA as output
              call      #delay
              or        outa, sclPin            ' Set SCL HIGH
              call      #delay
              andn      outa, sclPin            ' Set SCL LOW
              call      #delay
              mov       parm, data              ' Return the data byte
readbyte_ret  ret


delay         mov       delaycnt, cnt
              add       delaycnt, #DELAY_CYCLES
              waitcnt   delaycnt, #0
delay_ret     ret




sclPin                  long    0   ' 1 << I2C_SCL
sdaPin                  long    0   ' 1 << I2C_SDA

itgDLPF                 long    0

loopDelay               long    80_000_000 / 200
loopCount               long    0

i2cWordReadMask         long    %10000000_00000000




innerLoopCnt  res       1

accX          res       1
accY          res       1
accZ          res       1


drift         res       1
iT            res       1
divisor       res       1
dividend      res       1
tempOffset    res       1
signbit       res       1
divCounter    res       1
resultShifted res       1




hubAddr       res       1
devType       res       1
counter       res       1
updateCounter res       1
delaycnt      res       1
command       res       1
parm          res       1
data          res       1
ackbit        res       1

magCnt        res       1

inData        res       1
loByte        res       1
hiByte        res       1

readAddr      res       1
writeAddr     res       1

regAddr       res       1
regData       res       1

negativeOne   res       1